from cartographer.serializers import JSONAPISerializer
from cartographer.utils.version import JSONAPIVersion


class JSONAPICollectionSerializer(JSONAPISerializer):
    """This class be either be used directly or subclassed"""
    always_heterogeneous = False
    homogeneous_type = None
    links = {}

    def __init__(self, members=[], links={}):
        self._members = list(members) if members else members
        self.links = links

    def members(self):
        """Override in subclasses"""
        return self._members

    def is_collection(self):
        return True

    def next_page_url(self):
        return self.links.get('next')

    def prev_page_url(self):
        return self.links.get('prev')

    def last_page_url(self):
        return self.links.get('last')

    def first_page_url(self):
        return self.links.get('first')

    def pagination_urls(self):
        pagination_urls = {}
        if self.next_page_url():
            pagination_urls["next"] = self.next_page_url()
        if self.prev_page_url():
            pagination_urls["prev"] = self.prev_page_url()
        if self.first_page_url():
            pagination_urls["first"] = self.first_page_url()
        if self.last_page_url():
            pagination_urls["last"] = self.last_page_url()
        return pagination_urls

    def relationship_urls_json(self, version):
        relationship_urls_json = super().relationship_urls_json(version)
        relationship_urls_json.update(self.pagination_urls())
        return relationship_urls_json

    def as_link_json_rc2(self, version):
        resource_types = set([member.resource_type() for member in self.members()])
        is_heterogeneous = len(resource_types) > 1
        if is_heterogeneous or self.always_heterogeneous:
            data = [member.as_link_json() for member in self.members()]
            link_json = {
                "data": data
            }
            link_json.update(self.relationship_urls_json(version))
            return link_json
        resource_type = resource_types.pop() if resource_types else self.homogeneous_type
        if resource_type:
            link_json = {
                "type": resource_type,
                "ids": [member.resource_id_str() for member in self.members()]
            }
            link_json.update(self.relationship_urls_json(version))
            return link_json
        return {
            "ids": []
        }

    def as_linkage_json(self):
        return [member.as_linkage_json() for member in self.members()]

    def as_link_json(self, version):
        # if we drop RC2, then this override will probably go away
        if version == JSONAPIVersion.JSONAPI_RC2:
            return self.as_link_json_rc2(version)
        elif version == JSONAPIVersion.JSONAPI_RC3 or version == JSONAPIVersion.JSONAPI_1_0:
            key = 'data' if version == JSONAPIVersion.JSONAPI_1_0 else 'linkage'
            link_json = {key: self.as_linkage_json()}
            if version == JSONAPIVersion.JSONAPI_1_0:
                link_json['links'] = self.relationship_urls_json(version)
            else:
                link_json.update(self.relationship_urls_json(version))
            meta = self.meta()
            if meta is not None:
                link_json["meta"] = meta
            return link_json
        else:
            raise ValueError("Unknown JSON API version")

    def as_json_api_data(self, version):
        return [member.as_json_api_data(version) for member in self.members()]

    def list_of_linked_resources(self):
        return self.members()



from .schema_attribute import SchemaAttribute
from .schema_relationship import SchemaRelationship
from .array_relationship import ArrayRelationship
from .bool_attribute import BoolAttribute
from .date_attribute import DateAttribute
from .int_attribute import IntAttribute
from .json_attribute import JSONAttribute
from .string_attribute import StringAttribute



class SchemaAttribute(object):
    """
    `SchemaAttribute` describes how to translate attributes between JSON API and our Python models.

    `SchemaAttribute` is a very simple class, with two primary methods:
    `value`, for serializing from Python into JSON,
    and `from_json` for parsing from JSON into Python.
    Subclasses of `SchemaAttribute` can override either of these methods
    to customize (de)serialization behavior, typically per JSON output type
    (string, int, bool, date, etc.)
    """

    def __init__(self, model_attribute=None, is_property=True,
                 resource_method=None):
        """
        Note: only one of model_attribute and resource_method should be provided.
        Note: is_property is not relevant for resource_method

        :param model_attribute: the name of the attribute on the python model that should be (de)serialized
        :param is_property: whether or not model_attribute is a property (if False, it is presumed it is a method)
        :param resource_method: the name of the method on the resource object which uses this schema
        which should be called to get the serialized value.
        :return: an instance of SchemaAttribute,
        which will later be used by a Parser or Resource to move between JSON API and Python
        """

        identifier_args = [model_attribute, resource_method]
        provided_identifiers = [identifier
                                for identifier in identifier_args
                                if identifier]
        if len(provided_identifiers) > 1:
            raise Exception("only one of [{}] should be provided".format(identifier_args.join(", ")))

        self.model_attribute = model_attribute
        self.is_property = is_property
        self.resource_method = resource_method

    def value(self, resource):
        """
        :param resource: The resource which is currently responsible for serializing this value
        :return: The raw value to be serialized, which will be formatted via format_value
        """
        value = None
        if self.model_attribute:
            value = getattr(resource.model, self.model_attribute)
            if not self.is_property:
                value = value()
        elif self.resource_method:
            value = getattr(resource, self.resource_method)()

        if value is not None:
            value = self.format_value(value)

        return value

    @classmethod
    def format_value(cls, value):
        """
        :param value: The raw value, almost always the output of self.value
        :return: The formatted value, which can be dropped into a Python dictionary as the output of this Resource,
        typically which will have json.dumps called on it to return to a client
        """
        return value

    @classmethod
    def from_json(cls, serialized_value):
        """
        :param serialized_value: The value to be parsed, as pulled directly from RequestInterface.get_json(force=True)
        :return: The parsed value, which can be dropped into a Python dictionary as the output of this Parser,
        typically which will be inserted into a SQLAlchemy table.
        """
        return serialized_value



AUTHENTICATION_REQUIRED_401 = 401
NOT_AUTHORIZED_403 = 403
INVALID_REQUEST_400 = 400
REQUEST_ENTITY_TOO_LARGE_413 = 413
NOT_FOUND_404 = 404
INTERNAL_SERVER_ERROR_500 = 500
RESOURCE_CONFLICT_409 = 409
PAYMENT_REQUIRED_402 = 402
HEADER_PRECONDITIONS_FAILED = 412
LOCKED_423 = 423
TOO_MANY_REQUESTS_429 = 429


class JSONAPIException(Exception):
    status_code = INTERNAL_SERVER_ERROR_500
    error_title = "Internal Error."
    error_type_ = None
    error_description = None

    @property
    def error_type(self):
        return self.error_type_ or type(self).__name__

    def __init__(self, status_code=None, error_type=None, error_title=None,
                 error_description=None):
        self.status_code = status_code or self.status_code
        self.error_title = error_title or self.error_title
        self.error_description = error_description or self.error_description or self.error_title
        self.error_type_ = error_type


class ParameterMissing(JSONAPIException):
    status_code = INVALID_REQUEST_400
    parameter_name = None

    def __init__(self, parameter_name=None):
        self.error_title = \
            "Parameter '{}' is missing.".format(parameter_name or self.parameter_name)


class ParameterInvalid(JSONAPIException):
    status_code = INVALID_REQUEST_400

    def __init__(self, parameter_name, parameter_value):
        self.error_title = \
            "Invalid value for parameter '{}'.".format(parameter_name)
        self.error_description = \
            "Invalid parameter for '{0}': {1}.".format(parameter_name, parameter_value)


class BadPageCountParameter(ParameterInvalid):
    def __init__(self, parameter_value):
        super().__init__(parameter_name='page[count]',
                         parameter_value=parameter_value)
        self.error_description = "Page sizes must be integers."


class BadPageCursorParameter(ParameterInvalid):
    def __init__(self, parameter_value):
        super().__init__(parameter_name='page[cursor]',
                         parameter_value=parameter_value)
        self.error_description = "Provided cursor was not parsable."


class BadPageOffsetParameter(ParameterInvalid):
    def __init__(self, parameter_value):
        super().__init__(parameter_name='page[offset]',
                         parameter_value=parameter_value)
        self.error_description = "Page offsets must be integers."


class DataMissing(ParameterMissing):
    parameter_name = 'data'



class BaseMask(object):
    """
    Masks are the objects that are responsible for controlling user access
    to particular resources, and their attributes and relationships.

    To control access to the resources themselves, Masks have five methods, corresponding to CRUDL:
    * `can_create` (true by default)
    * `can_view` (true by default)
    * `can_list` (false by default)
    * `can_edit` (false by default)
    * `can_delete` (false by default)

    These can and should be used in your route file / controller to control access to the requested resource(s).

    Should one be allowed access at that level,
    Masks can provide more fine-grained access control at the per-attribute and per-relationship level:
    * `fields_cant_view` (always called after `can_view`)
    * `fields_cant_edit` (always called after `can_edit`)
    * `includes_cant_create` (always called after `can_edit`)
    * `includes_cant_view` (always called after `can_view`)
    * `includes_cant_edit` (always called after `can_edit`)
    * `includes_cant_delete` (always called after `can_edit`)

    By using `SchemaResource` and `SchemaParser`,
    the corresponding Mask for your resource will be used at (de)serialization time
    to appropriately remove fields from the output, or disallow their input.
    `SchemaResource` and `SchemaParser`, however,
    will assume that the CRUDL checks have already been used,
    and will not check whether or not the resource *itself* should be (de)serialized.
    As such, please remember to use the CRUDL control methods in your routes.
    """

    @classmethod
    def can_create(cls, user_id):
        """
        Typically used in routes like `POST /widgets`

        :param user_id: The user on behalf of whom permission is being requested
        :return: A boolean value indicating whether or not the given user
        is allowed to create wholly new instances of this resource
        """
        return True

    @classmethod
    def can_view(cls, model, user_id):
        """
        Typically used in routes like `GET /widgets/12`
        Note: Sometimes the user is allowed to see a limited version of the requested resource.
        If that is the case, this method should return True,
        and `fields_cant_view` and `includes_cant_view` should be used
        to control which attributes and relationships are actually visible

        :param model: The model which the user is trying to view
        :param user_id: The user on behalf of whom permission is being requested
        :return: A boolean value indicating whether or not the given user
        is allowed to view the given resource at all.
        """
        return True

    @classmethod
    def can_list(cls, user_id):
        """
        Typically used in routes like `GET /widgets`

        :param user_id: The user on behalf of whom permission is being requested
        :return: A boolean value indicating whether or not the given user
        is allowed to list all resources of this type.
        """
        return False

    @classmethod
    def can_edit(cls, model, user_id):
        """
        Typically used in routes like `PATCH /widgets/12`
        Note: Sometimes the user is allowed to edit a limited version of the requested resource.
        If that is the case, this method should return True,
        and `fields_can_edit` and `includes_cant_edit` should be used
        to control which attributes and relationships are actually editable.

        :param model: The model which the user is trying to edit
        :param user_id: The user on behalf of whom permission is being requested
        :return: A boolean value indicating whether or not the given user
        is allowed to edit any attributes or relationships of the given resource.
        """
        return False

    @classmethod
    def can_delete(cls, model, user_id):
        """
        Typically used in routes like `DELETE /widgets/12`

        :param model: The model which the user is trying to delete
        :param user_id: The user on behalf of whom permission is being requested
        :return: A boolean value indicating whether or not the given user
        is allowed to delete the given resource.
        """
        return False

    @classmethod
    def fields_cant_view(cls, model, user_id):
        """
        :param model: The model which the user is trying to view attributes of
        :param user_id: The user on behalf of whom permission is being requested
        :return: A list of strings indicating which attributes the given user is not allowed to view.
        """
        return []

    @classmethod
    def fields_cant_edit(cls, model, user_id):
        """
        :param model: The model which the user is trying to edit
        :param user_id: The user on behalf of whom permission is being requested
        :return: A list of strings indicating which attributes the given user is not allowed to edit.
        """
        return []

    @classmethod
    def includes_cant_create(cls, model, user_id):
        """
        :param model: The model which the user is trying to edit by creating new relationships
        :param user_id: The user on behalf of whom permission is being requested
        :return: A list of strings indicating which relationships the given user
        is not allowed to create or add to.
        """
        return []

    @classmethod
    def includes_cant_view(cls, model, user_id):
        """
        :param model: The model which the user is trying to view relationships of
        :param user_id: The user on behalf of whom permission is being requested
        :return: A list of strings indicating which relationships the given user is not allowed to view.
        """
        return []

    @classmethod
    def includes_cant_edit(cls, model, user_id):
        """
        :param model: The model which the user is trying to edit
        :param user_id: The user on behalf of whom permission is being requested
        :return: A list of strings indicating which relationships the given user is not allowed to edit.
        """
        return []

    @classmethod
    def includes_cant_delete(cls, model, user_id):
        """
        :param model: The model which the user is trying to edit by deleting existing relationships
        :param user_id: The user on behalf of whom permission is being requested
        :return: A list of strings indicating which relationships the given user
        is not allowed to delete or remove from.
        """
        return []



import ciso8601

import dateutil.parser

from cartographer.field_types import SchemaAttribute
from cartographer.utils.datetime import as_utc, make_naive


class DateAttribute(SchemaAttribute):
    @classmethod
    def format_value(cls, value):
        return as_utc(value).isoformat()

    @classmethod
    def from_json(cls, serialized_value):
        try:
            parsed_value = ciso8601.parse_datetime(serialized_value)
            assert parsed_value is not None  # Caveat: asserts won't run if python is run with -O.
        except Exception as e:
            parsed_value = dateutil.parser.parse(serialized_value)
        return make_naive(parsed_value)



import importlib
from collections import deque

from cartographer.utils.version import get_default_version, JSONAPIVersion


class JSONAPISerializer(object):
    def resource_id(self):
        """Override this in a subclass to return a unique id string"""
        raise NotImplementedError()

    @classmethod
    def resource_type(cls):
        """Override this in a subclass to return a type name string"""
        raise NotImplementedError()

    def as_json(self):
        """Override this in a subclass to return json"""
        raise NotImplementedError()

    def linked_resources(self):
        """Override this in a subclass to link to other resources"""
        return {}

    def resource_url(self):
        return None

    def relationship_url(self):
        return None

    # internal usage

    def as_json_api_document(self, version=None):
        if version is None:
            version = self._flask_json_api_version()
        if version is None:
            version = get_default_version()
        return self.document_with_data(self.as_json_api_data(version), version)

    def resource_id_str(self):
        id_ = self.resource_id()
        return str(id_) if id_ is not None else None

    def resource_key(self):
        try:
            if self.resource_id_str():
                return (self.resource_type(), self.resource_id_str())
            else:  # Collection or Null
                return None
        except NotImplementedError:
            return None

    def is_collection(self):
        return False

    def list_of_linked_resources(self):
        return self.linked_resources().values()

    def deeply_linked_resources(self, skip_self=True):
        all_linked_resources = []
        q = deque()
        resources_enqueued = set()

        def enqueue_if_never_enqued(linked_resource):
            linked_key = linked_resource.resource_key()
            if (not linked_key or
                    (linked_key not in resources_enqueued)):
                q.append(linked_resource)
                resources_enqueued.add(linked_key)

        def enqueue_linked_resources(resource):
            for linked_resource in resource.list_of_linked_resources():
                enqueue_if_never_enqued(linked_resource)

        if skip_self:  # prevent the toplevel object/collection from appearing in "linked"
            if self.is_collection(): # self is a collection, skip its members
                # I'm iterating over self.members() twice to make sure that
                # everything in the top-level collection gets marked as enqueued
                # before we enqueue any of their linked resources.
                # this prevents links between top-level resources from
                # causing problems
                for member_resource in self.members():
                    resources_enqueued.add(member_resource.resource_key())
                for member_resource in self.members():
                    enqueue_linked_resources(member_resource)
            else: # self is a single object, skip it and enqueue its children
                resources_enqueued.add(self.resource_key())
                enqueue_linked_resources(self)
        else:
            enqueue_if_never_enqued(self)

        while q:
            resource = q.popleft()
            if resource.resource_key():
                all_linked_resources.append(resource)

            enqueue_linked_resources(resource)

        return all_linked_resources

    def deeply_linked_resources_as_json(self, version, skip_self=True):
        return list(filter(None, [
            resource.as_json_api_data(version)
            for resource
            in self.deeply_linked_resources(skip_self)
        ]))

    def resource_links_json(self, version):
        links_json = {
            link_name: resource.as_link_json(version)
            for link_name, resource in self.linked_resources().items()
        }
        if self.resource_url():
            links_json["self"] = self.resource_url()
        return links_json

    def as_linkage_json(self):
        return {"type": self.resource_type(), "id": self.resource_id_str()}

    def relationship_urls_json(self, version):
        relationship_urls_json = {}
        if self.relationship_url():
            relationship_urls_json["self"] = self.relationship_url()
        if self.resource_url():
            if version == JSONAPIVersion.JSONAPI_RC2:
                relationship_urls_json["resource"] = self.resource_url()
            elif version == JSONAPIVersion.JSONAPI_RC3 or version == JSONAPIVersion.JSONAPI_1_0:
                relationship_urls_json["related"] = self.resource_url()
            else:
                raise ValueError("Unknown JSON API version")
        return relationship_urls_json

    def as_link_json(self, version):
        link_json = {}

        if version == JSONAPIVersion.JSONAPI_RC2:
            link_json.update(self.as_linkage_json())
        elif version == JSONAPIVersion.JSONAPI_RC3:
            link_json["linkage"] = self.as_linkage_json()
        elif version == JSONAPIVersion.JSONAPI_1_0:
            link_json["data"] = self.as_linkage_json()
        else:
            raise ValueError("Unknown JSON API version")

        link_json.update(self.relationship_urls_json(version))
        meta = self.meta()
        if meta is not None:
            link_json["meta"] = meta

        return link_json

    def as_json_api_data(self, version):
        if version == JSONAPIVersion.JSONAPI_1_0:
            attributes = self.as_json()
            if 'id' in attributes:
                del attributes['id']
            json = {'attributes': attributes}
        else:
            json = self.as_json()
        json["id"] = self.resource_id_str()
        json["type"] = self.resource_type()
        links = self.resource_links_json(version)
        if links:
            if version == JSONAPIVersion.JSONAPI_1_0:
                json["relationships"] = links
            else:
                json["links"] = links
        # TODO: optional keys: "meta"
        return json

    def next_page_url(self):
        return None

    def prev_page_url(self):
        return None

    def last_page_url(self):
        return None

    def first_page_url(self):
        return None

    def meta(self):
        return None

    def document_links_urls(self):
        links = {
            "next": self.next_page_url(),
            "prev": self.prev_page_url(),
            "last": self.last_page_url(),
            "first": self.first_page_url()
        }
        links = {key: value for key, value in links.items() if value}
        return links

    def included_resources_key(self, version):
        if version == JSONAPIVersion.JSONAPI_RC2:
            return "linked"
        elif version == JSONAPIVersion.JSONAPI_RC3 or version == JSONAPIVersion.JSONAPI_1_0:
            return "included"
        else:
            raise ValueError("Unknown JSON API version")

    def as_json_api_relationship_document(self, version=None):
        if version is None:
            version = self._flask_json_api_version()
        if version is None:
            version = get_default_version()
        return self.document_with_data(self.as_linkage_json(), version, False)

    def document_with_data(self, data, version, skip_self=True):
        response = {"data": data}
        included_resources = self.deeply_linked_resources_as_json(version, skip_self)
        if included_resources:
            response[self.included_resources_key(version)] = included_resources
        links = self.document_links_urls()
        if links:
            response["links"] = links
        meta = self.meta()
        if meta is not None:
            response["meta"] = meta
        return response

    # flask request context

    @staticmethod
    def _has_flask_request_context():
        return importlib.util.find_spec("flask") is not None

    @staticmethod
    def _flask_json_api_version():
        if JSONAPISerializer._has_flask_request_context():
            from flask import request
            if request:
                return request.get_json_api_version()
        return None



from cartographer.field_types import SchemaRelationship
from cartographer.resources.resource_registry import ResourceRegistryKeys


class ArrayRelationship(SchemaRelationship):
    def resource(self, parent_resource, relationship_key):
        from cartographer.serializers import JSONAPICollectionSerializer

        if self.resource_method is not None:
            return getattr(parent_resource, self.resource_method)()

        models = []
        if self.id_attribute is not None:
            # TODO: 'id' on to-many relations?
            models = []
        elif self.model_property is not None:
            models = getattr(parent_resource.model, self.model_property)
        elif self.model_method is not None:
            models = getattr(parent_resource.model, self.model_method)()

        resource_class = self.resource_registry_entry().get(ResourceRegistryKeys.SERIALIZER)
        # TODO: custom collection class, custom arguments to resource_class
        return JSONAPICollectionSerializer([
            resource_class(
                model,
                parent_resource=parent_resource,
                relationship_name=relationship_key
            )
            for model in models
        ])



from .jsonapi_serializer import JSONAPISerializer
from .jsonapi_null_serializer import JSONAPINullSerializer
from .jsonapi_collection_serializer import JSONAPICollectionSerializer
from .schema_serializer import SchemaSerializer



from .resource_registry import ResourceRegistry

_resource_registry = ResourceRegistry()


def get_resource_registry(registry=_resource_registry):
    registry.call_initialization_hook()
    return registry.registry



from enum import Enum

from cartographer.utils.collections import filter_dict


class ResourceRegistryKeys(Enum):
    TYPE = "type"
    SCHEMA = "schema"
    MODEL = "model"
    SERIALIZER = "serializer"
    PARSER = "parser"
    MASK = "mask"


class ResourceRegistry(object):
    def __init__(self):
        self.registry = {}
        self.initialization_hook = None
        self.hook_called = False

    def set_initialization_hook(self, hook=None):
        self.initialization_hook = hook

    def call_initialization_hook(self):
        if not self.hook_called and self.initialization_hook:
            self.initialization_hook()
            self.hook_called = True

    def register_resource(self, type_string, schema, model=None,
                          serializer=None, parser=None, mask=None):
        self.registry[type_string] = filter_dict({
            ResourceRegistryKeys.TYPE: type_string,
            ResourceRegistryKeys.SCHEMA: schema,
            ResourceRegistryKeys.MODEL: model,
            ResourceRegistryKeys.SERIALIZER: serializer,
            ResourceRegistryKeys.PARSER: parser,
            ResourceRegistryKeys.MASK: mask,
        })



class Schema(object):
    """
    The core element of the Cartographer system is the Schema.
    The schema is a map from our models to their API output and vice versa.
    Subclasses should define a class variable SCHEMA,
    formatted to resemble the output structure of JSON API resources:

    ```python
    {
        "type": "widget",
        "id": StringAttribute("widget_id"),
        "atributes": {
            "price": IntAttribute("amount_cents")
        },
        "relationships": {
            "manufacturer": SchemaRelationship(
                model_type="manufacturer",
                id_attribute="manufacturer_id"
            )
        }
    }
    ```

    Each of these top-level keys has corresponding utility methods on this class:
    * `resource_type` to access the value of `type` in the schema
    * `resource_id` to access the `SchemaAttribute` class corresponding to the `id` of the resource
    * `attributes` to list all attribute keys, and `attribute` to look up one particular attribute
    * `relationships` to list all attribute keys, and `relationship` to look up one particular relationship
    """

    @classmethod
    def schema(cls):
        """Override this or provide a class variable SCHEMA in a subclass to define model <=> API mappings"""
        return cls.SCHEMA

    @classmethod
    def route_prefix(cls):
        """Override this to define the API route namespace for these resources"""
        # TODO: migrate this to `route_prefixes`, and [cls.resource_type()] is the default,
        # with subclasses appending the plural form if they want.
        return None

    # Convenience methods for accessing pieces of the schema.

    @classmethod
    def resource_type(cls):
        return cls.schema().get('type')

    @classmethod
    def resource_id(cls):
        return cls.schema().get('id')

    @classmethod
    def attributes(cls):
        return list(cls.schema().get('attributes', {}).keys())

    @classmethod
    def attribute(cls, key):
        return cls.schema().get('attributes', {}).get(key)

    @classmethod
    def relationships(cls):
        return list(cls.schema().get('relationships', {}).keys())

    @classmethod
    def relationship(cls, key):
        return cls.schema().get('relationships', {}).get(key)

    @classmethod
    def schema_lookup(cls, json_key):
        return cls.schema().get(json_key)



from setuptools import setup, find_packages

setup(name='cartographer',
      version='0.0.2',
      description='Python library for using JSON API with Flask.',
      url='http://github.com/Patreon/cartographer',
      author='Patreon',
      author_email='david@patreon.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'python-dateutil==2.4.2',
          'ciso8601==1.0.1'
      ],
      zip_safe=True)



import itertools

from cartographer.utils.version import get_default_version


class PostedDocument(object):
    """This is a reader for JSON API Documents"""
    def __init__(self, json_data, version=None):
        self.json_data = json_data

        if version is None:
            version = get_default_version()
        self.version = version

    def data(self):
        if isinstance(self.json_data["data"], list):
            return [PostedResource(datum, self) for datum in self.json_data["data"]]
        else:
            return PostedResource(self.json_data["data"], self)

    def all_resource_json_data(self):
        resources = []
        if isinstance(self.json_data["data"], list):
            resources = itertools.chain(resources, self.json_data["data"])
        else:
            resources = itertools.chain(resources, [self.json_data["data"]])

        if "included" in self.json_data:
            resources = itertools.chain(resources, self.json_data["included"])

        return resources

    def find_resource_by_type_and_id(self, resource_type, resource_id):
        for resource_json_data in self.all_resource_json_data():
            if resource_json_data.get("type") == resource_type and resource_json_data.get("id") == resource_id:
                return PostedResource(resource_json_data, self)
        return None


class PostedResource(object):
    """Represents a single object in a JSON API Document"""
    def __init__(self, json_data, document):
        self.json_data = json_data
        self.document = document

    def resource_type(self):
        return self.json_data["type"]

    def attribute(self, name):
        return self.json_data["attributes"].get(name)

    def assert_type(self, resource_type):
        if self.resource_type() != resource_type:
            raise Exception("Expected a " + resource_type + ", but got a " + self.resource_type())
        return self

    def relationship(self, relationship_name):
        if "relationships" not in self.json_data:
            return None

        if relationship_name not in self.json_data["relationships"]:
            return None

        return PostedRelationship(self.json_data["relationships"][relationship_name], self.document)

    def relationship_id(self, relationship_name):
        relationship = self.relationship(relationship_name)
        if relationship:
            return relationship.relationship_id()
        else:
            return None

    def related_resource(self, relationship_name):
        relationship = self.relationship(relationship_name)
        if relationship:
            return relationship.resource()
        else:
            return None


class PostedRelationship(object):
    """Represents a named relationship in a JSON API Document"""
    def __init__(self, json_data, document):
        self.json_data = json_data
        self.document = document

    def relationship_id(self):
        if "data" in self.json_data:
            if isinstance(self.json_data["data"], list):
                return [PostedRelationshipID(datum, self.document) for datum in self.json_data["data"]]
            else:
                return PostedRelationshipID(self.json_data["data"], self.document)

    def resource(self):
        relationship_id = self.relationship_id()

        if isinstance(relationship_id, list):
            return [one_id.related_resource() for one_id in relationship_id]
        else:
            return relationship_id.related_resource()


class PostedRelationshipID(object):
    """Represents a the actual data of a relationship in a JSON API Document"""
    def __init__(self, json_data, document):
        self.json_data = json_data
        self.document = document

    def resource_type(self):
        return self.json_data.get("type")

    def resource_id(self):
        return self.json_data.get("id")

    def assert_type(self, resource_type):
        if self.resource_type() != resource_type:
            raise Exception("Expected a " + resource_type + ", but got a " + self.resource_type())
        return self

    def related_resource(self):
        return self.document.find_resource_by_type_and_id(self.resource_type(), self.resource_id())
